from __future__ import annotations

from pathlib import Path

import pyarrow.parquet as pq

__all__ = ["AssemblyInfo"]


class AssemblyInfo:
    _instance = None
    _db_path = Path(__file__).parent / "data" / "db.parquet"

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._load_db()
        return cls._instance

    def _load_db(self) -> None:
        """Private method to connect to the database."""
        self._data = pq.read_table(self._db_path).to_pandas()

    @classmethod
    def connect(cls):
        """Returns the singleton instance of AssemblyInfo."""
        return cls()



from assemblyinfo.interface import AssemblyInfo


def test_method_attachment():
    db = AssemblyInfo.connect()
    assert hasattr(db, "info"), "Method info is not attached"


def test_method_execution():
    genome_info = AssemblyInfo.connect()

    genome_info.info()


def test_get_species_info():
    genome_info = AssemblyInfo.connect()

    assert hasattr(
        genome_info, "get_species_info"
    ), "Method get_species_info is not attached"

    rs = genome_info.get_species_info("homo_sapiens")

    assert rs is not None, "The result is None"



def test_get_assembly_metadata_ncbi():
    genome_info = AssemblyInfo.connect()

    assert hasattr(
        genome_info, "get_assembly_metadata"
    ), "Method get_assembly_metadata is not attached"

    rs = genome_info.get_assembly_metadata("GRCh38")

    assert rs is not None, "The result is None"



def test_get_assembly_metadata_ucsc():
    genome_info = AssemblyInfo.connect()

    rs = genome_info.get_assembly_metadata("hg38")


    assert rs is not None, "The result is None"
    assert len(rs) > 0, "The result if empty"


def test_get_assembly_metadata_t2t():
    genome_info = AssemblyInfo.connect()

    rs = genome_info.get_assembly_metadata("T2T-CHM13")

    assert rs is not None, "The result is None"
    assert len(rs) > 0, "The result if empty"


def test_available_assemblies():
    genome_info = AssemblyInfo.connect()

    rs = genome_info.available_assemblies()

    assert rs is not None, "The result is None"
    assert len(rs) > 0, "The result is empty"


def test_get_organism_info():
    genome_info = AssemblyInfo.connect()

    rs = genome_info.get_organism_info("human")

    assert rs is not None, "The result is None"
    assert len(rs) > 0, "The result is empty"


def test_available_patches():
    genome_info = AssemblyInfo.connect()

    rs = genome_info.available_patches()

    assert rs is not None, "The result is None"
    assert len(rs) > 0, "The result is empty"


def test_available_accessions():
    genome_info = AssemblyInfo.connect()

    rs = genome_info.available_accessions("GRCh38")

    assert rs is not None, "The result is None"
    assert len(rs) > 0, "The result is empty"

    rs = genome_info.available_accessions("hg38")

    assert rs is not None, "The result is None"
    assert len(rs) > 0, "The result is empty"



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from datetime import datetime
from importlib.metadata import metadata

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

info = metadata("assemblyinfo")
project_name = info["Name"]
project = "AssemblyInfo"
author = "Open2C"
copyright = f"{datetime.now():%Y}, {author}."
version = info["Version"]
urls = dict(pu.split(", ") for pu in info.get_all("Project-URL"))

# The full version, including alpha/beta/rc tags
release = info["Version"]

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",  # 'numpydoc'
    "myst_parser",
    "sphinx_rtd_theme",
    "sphinx.ext.autodoc",
    "autodocsumm",
    "nbsphinx",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



from __future__ import annotations

import pandas as pd

__all__ = [
    "filter_chromosome_data",
    "get_chromnames",
    "get_chromsizes",
    "get_chrom_eq",
    "get_seqinfo",
]


def filter_chromosome_data(
    cls,
    assembly: str,
    roles: list[str] | None = None,
    units: list[str] | None = None,
    length: str | None = None,
) -> pd.DataFrame:
    """
    Filters the chromosome data based on the provided parameters.

    Parameters
    ----------
    assembly : str
        The assembly name to filter by.
    roles : list[str], optional
        The roles to filter by.
    units : list[str], optional
        The units to filter by.
    length : Optional[str]
        The length condition to filter by (e.g., '> 1000').

    Returns
    -------
    pd.DataFrame
        The filtered chromosome data.

    Raises
    ------
    ValueError
        If the assembly is not found in the database.

    Examples
    --------
    >>> AssemblyInfo.filter_chromosome_data("hg38", roles=["assembled"])
    """
    if assembly in cls._data["assembly"].tolist():
        group = "assembly"
    elif assembly in cls._data["assembly_ucsc"].dropna().tolist():
        group = "assembly_ucsc"
    else:
        raise ValueError(f"{assembly} not in database!")

    q1 = f'{group} == "{assembly}" and version == "latest"'
    q2 = ""

    if length:
        q2 += f"length {length}"

    if roles:
        if len(q2) > 0:
            q2 += f" and role.isin({roles})"
        else:
            q2 += f"role.isin({roles})"

    if units:
        if len(q2) > 0:
            q2 += f"  and unit.isin({units})"
        else:
            q2 += f"unit.isin({units})"

    if len(q2) > 0:
        mask = (
            pd.DataFrame.from_records(cls._data.query(q1).seqinfo.iloc[0])
            .query(q2)
            .convert_dtypes()
        )
    else:
        mask = pd.DataFrame.from_records(
            cls._data.query(q1).seqinfo.iloc[0]
        ).convert_dtypes()

    return mask


def get_chromnames(
    cls,
    assembly: str,
    provider: str | None = None,
    roles: list[str] | None = None,
    units: list[str] | None = None,
    length: str | None = None,
) -> list[str]:
    """
    Returns the chromosome names for the specified assembly.

    Parameters
    ----------
    assembly : str
        The assembly name to filter by.
    provider : Optional[str]
        The provider to filter by ('ucsc', 'genbank', 'refseq', 'ncbi').
    roles : Optional[List[str]]
        The roles to filter by.
    units : Optional[List[str]]
        The units to filter by.
    length : Optional[str]
        The length condition to filter by (e.g., '> 1000').

    Returns
    -------
    List[str]
        A list of chromosome names.

    Raises
    ------
    ValueError
        If the provider is not valid.

    Examples
    --------
    >>> AssemblyInfo.get_chromnames("hg38", provider="ucsc")
    """
    if not provider or provider == "ucsc":
        colname = "name"
    elif provider in ["genbank", "refseq", "ncbi"]:
        colname = provider
    else:
        error_msg = (
            f"{provider} is not a valid provider!\n",
            "Valid providers are 'ucsc', 'genbank', 'refseq', 'ncbi'",
        )
        raise ValueError(error_msg)

    return cls.filter_chromosome_data(assembly, roles, units, length)[colname].tolist()


def get_chromsizes(
    cls,
    assembly: str,
    provider: str | None = None,
    roles: list[str] | None = None,
    units: list[str] | None = None,
    length: str | None = None,
) -> pd.Series:
    """
    Returns the chromosome sizes for the specified assembly.

    Parameters
    ----------
    assembly : str
        The assembly name to filter by.
    provider : Optional[str]
        The provider to filter by ('ucsc', 'genbank', 'refseq', 'ncbi').
    roles : Optional[List[str]]
        The roles to filter by.
    units : Optional[List[str]]
        The units to filter by.
    length : Optional[str]
        The length condition to filter by (e.g., '> 1000').

    Returns
    -------
    pd.Series
        A series with chromosome sizes.

    Raises
    ------
    ValueError
        If the provider is not valid.

    Examples
    --------
    >>> AssemblyInfo.get_chromsizes("hg38", provider="ucsc")
    """
    if not provider or provider == "ucsc":
        colname = "name"
    elif provider in ["genbank", "refseq", "ncbi"]:
        colname = provider
    else:
        error_msg = (
            f"{provider} is not a valid provider!\n",
            "Valid providers are 'ucsc', 'genbank', 'refseq', 'ncbi'",
        )
        raise ValueError(error_msg)

    df = cls.filter_chromosome_data(assembly, roles, units, length)
    return df.set_index(colname)["length"]


def get_chrom_eq(
    cls,
    assembly: str,
    providers: list[str] | None = None,
    roles: list[str] | None = None,
    units: list[str] | None = None,
    length: str | None = None,
) -> pd.DataFrame:
    """
    Returns the chromosome equivalence for the specified assembly.

    Parameters
    ----------
    assembly : str
        The assembly name to filter by.
    providers : Optional[List[str]]
        The providers to filter by (e.g., 'ucsc', 'genbank', 'refseq', 'ncbi').
    roles : Optional[List[str]]
        The roles to filter by.
    units : Optional[List[str]]
        The units to filter by.
    length : Optional[str]
        The length condition to filter by (e.g., '> 1000').

    Returns
    -------
    pd.DataFrame
        A DataFrame with chromosome equivalence across different providers.

    Raises
    ------
    ValueError
        If the provider is not valid.

    Examples
    --------
    >>> AssemblyInfo.get_chrom_eq("hg38", providers=["ucsc", "genbank"])
    """
    if not providers:
        providers = ["name", "ncbi", "genbank", "refseq"]
    elif "ucsc" in providers:
        providers = [p if p != "ucsc" else "name" for p in providers]

    return cls.filter_chromosome_data(assembly, roles, units, length)[providers]


def get_seqinfo(cls, assembly: str) -> pd.DataFrame:
    """
    Returns the sequence information for the specified assembly.

    Parameters
    ----------
    assembly : str
        The assembly name to filter by.

    Returns
    -------
    pd.DataFrame
        A DataFrame with sequence information.

    Raises
    ------
    ValueError
        If the assembly is not found in the database.

    Examples
    --------
    >>> AssemblyInfo.get_seqinfo("hg38")
    """
    if assembly in cls._data["assembly"].tolist():
        group = "assembly"
    elif assembly in cls._data["assembly_ucsc"].dropna().tolist():
        group = "assembly_ucsc"
    elif assembly in cls._data["patch"].dropna().tolist():
        group = "patch"
    else:
        error_msg = (
            f"{assembly} not in database!\n",
            "Valid assemblies are:\n\n",
            f"NCBI:\n{cls._data.assembly.unique().tolist()}\n\n",
            f"UCSC:\n{cls._data.assembly_ucsc.dropna().unique().tolist()}\n\n",
            f"Patch:\n{cls._data.patch.dropna().unique().tolist()}",
        )
        raise ValueError(error_msg)

    q1 = f'{group} == "{assembly}"'
    q2 = 'version == "latest"'

    local_db = cls._data.query(q1)
    if len(local_db) > 1:
        return (
            pd.DataFrame.from_records(local_db.query(q2).seqinfo.iloc[0])
            .convert_dtypes()
            .set_index("name")
        )
    else:
        return (
            pd.DataFrame.from_records(local_db.seqinfo.iloc[0])
            .convert_dtypes()
            .set_index("name")
        )



import pytest

from assemblyinfo.interface import AssemblyInfo


def test_get_genbank_accession():
    db = AssemblyInfo.connect()

    result = db.get_genbank_accession("GRCh38.p14")
    assert result == ["GCA_000001405.29"]

    result = db.get_genbank_accession("GRCm38.p5")
    assert result == ["GCA_000001635.7"]

    with pytest.raises(ValueError):
        db.get_genbank_accession("NonExistentAssembly")


def test_get_refseq_accession():
    db = AssemblyInfo.connect()

    result = db.get_refseq_accession("GRCh38.p14")
    assert result == ["GCF_000001405.40"]

    result = db.get_refseq_accession("GRCm38.p5")
    assert result == ["GCF_000001635.25"]

    with pytest.raises(ValueError):
        db.get_refseq_accession("NonExistentAssembly")


def test_get_patch_from_accession():
    db = AssemblyInfo.connect()

    result = db.get_patch_from_accession("GCF_000001405.40")
    assert result == ["GRCh38.p14"]

    result = db.get_patch_from_accession("GCA_000001635.7")
    assert result == ["GRCm38.p5"]

    with pytest.raises(ValueError):
        db.get_patch_from_accession("NonExistentAssembly")


def test_get_assembly_from_accession():
    db = AssemblyInfo.connect()

    result = db.get_assembly_from_accession("GCA_000001405.29")
    assert result == ["GRCh38", "hg38"]

    result = db.get_assembly_from_accession("GCF_000001635.25")
    assert result == ["GRCm38", "mm10"]

    with pytest.raises(ValueError):
        db.get_assembly_from_accession("NonExistentAssembly")



from .acc import (
    get_assembly_from_accession,
    get_genbank_accession,
    get_patch_from_accession,
    get_refseq_accession,
)
from .assembly import (
    Assembly,
    assembly_info,
)
from .chrom import (
    filter_chromosome_data,
    get_chrom_eq,
    get_chromnames,
    get_chromsizes,
    get_seqinfo,
)
from .info import (
    available_accessions,
    available_assemblies,
    available_patches,
    available_species,
    build_assembly_info,
    get_assembly_metadata,
    get_db,
    get_info,
    get_organism_info,
    get_species_info,
    get_version,
    info,
)

__all__ = [
    "info",
    "get_db",
    "get_info",
    "get_species_info",
    "get_organism_info",
    "get_version",
    "build_assembly_info",
    "get_assembly_metadata",
    "available_assemblies",
    "available_patches",
    "available_species",
    "available_accessions",
    "get_genbank_accession",
    "get_refseq_accession",
    "get_patch_from_accession",
    "get_assembly_from_accession",
    "filter_chromosome_data",
    "get_chromnames",
    "get_chromsizes",
    "get_chrom_eq",
    "get_seqinfo",
    "Assembly",
    "assembly_info",
]



from dataclasses import dataclass
from typing import Dict, List, Optional

import pandas as pd

from .chrom import filter_chromosome_data
from .info import get_assembly_metadata

__all__ = ["Assembly", "assembly_info"]


@dataclass
class Assembly:
    """
    A dataclass to store assembly information.
    """

    assembly: str
    species: str
    common_name: str
    seqinfo: pd.DataFrame
    metadata: Dict[str, str]
    aliases: Dict[str, Dict[str, str]]

    @property
    def chromnames(self) -> List[str]:
        return self.seqinfo.index.tolist()

    @property
    def chromsizes(self) -> pd.Series:
        return self.seqinfo["length"]

    @property
    def chromeq(self) -> Dict[str, Dict[str, str]]:
        return pd.DataFrame(self.aliases).T

    def __repr__(self):
        return (f"Assembly(assembly={self.assembly}",
                f"species={self.species}",
                f"common_name={self.common_name})")


def assembly_info(
    cls,
    assembly: str,
    provider: Optional[str] = None,
    roles: Optional[List[str]] = None,
    units: Optional[List[str]] = None,
    length: Optional[str] = None,
) -> Assembly:
    """
    Get assembly information for a given assembly.
    Parameters
    ----------
    assembly : str
        Assembly name.
    provider : str, Optional
        Data provider.
    roles : List[str], Optional
        Chromosome roles.
    units : List[str], Optional
        Chromosome units.
    length : str, Optional
        Chromosome length.
    Returns
    -------
    Assembly
        Assembly information.
    Examples
    --------
    >>> assembly_info("hg38")
    >>> assembly_info("hg38", provider="ucsc")
    >>> assembly_info("hg38", roles=["assembled"])
    >>> assembly_info("hg38", units=["primary"])
    >>> assembly_info("hg38", length=">1000")
    """
    if provider == "ucsc" or provider is None:
        provider = "name"

    if provider not in ["name", "ncbi", "genbank", "refseq"]:
        raise ValueError("Invalid provider.")

    seqinfo = filter_chromosome_data(
        cls, assembly=assembly, roles=roles, units=units, length=length
    )

    aliases = (
        seqinfo[["name", "ncbi", "genbank", "refseq"]]
        .set_index(provider)
        .to_dict(orient="index")
    )

    metadata = get_assembly_metadata(cls, assembly=assembly)

    return Assembly(
        assembly=assembly,
        species=metadata["species"],
        common_name=metadata["common_name"],
        seqinfo=seqinfo.set_index(provider),
        metadata=metadata,
        aliases=aliases,
    )



import pandas as pd
import pytest

from assemblyinfo.interface import AssemblyInfo


def test_filter_chromosome_data():
    db = AssemblyInfo.connect()

    result = db.filter_chromosome_data("GRCh38", roles=["assembled"])
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 25
    assert set(result["role"]) == {"assembled"}

    result = db.filter_chromosome_data(
        "hg38", roles=["assembled"], units=["non-nuclear"]
    )
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 1
    assert set(result["role"]) == {"assembled"}
    assert set(result["unit"]) == {"non-nuclear"}

    result = db.filter_chromosome_data(
        "GRCh38", roles=["assembled"], length=">133137821"
    )
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 13
    assert set(result["role"]) == {"assembled"}

    with pytest.raises(ValueError):
        db.filter_chromosome_data("NonExistentAssembly")

    result = db.filter_chromosome_data("GRCh38", roles=["non_existent_role"])
    assert len(result) == 0


def test_get_chromnames():
    db = AssemblyInfo.connect()

    result = db.get_chromnames("GRCh38", roles=["assembled"])
    assert isinstance(result, list)
    assert len(result) == 25

    result = db.get_chromnames("hg38", roles=["assembled"], units=["non-nuclear"])
    assert isinstance(result, list)
    assert len(result) == 1

    with pytest.raises(ValueError):
        db.get_chromnames("NonExistentAssembly")


def test_get_chromsizes():
    db = AssemblyInfo.connect()

    result = db.get_chromsizes("GRCh38", roles=["assembled"])
    assert isinstance(result, pd.Series)
    assert len(result) == 25

    result = db.get_chromsizes("hg38", roles=["assembled"], units=["non-nuclear"])
    assert isinstance(result, pd.Series)
    assert len(result) == 1

    with pytest.raises(ValueError):
        db.get_chromsizes("NonExistentAssembly")


def test_get_seqinfo():
    db = AssemblyInfo.connect()

    result = db.get_seqinfo("GRCh38")
    assert isinstance(result, pd.DataFrame)

    result = db.get_seqinfo("T2T-CHM13")
    assert isinstance(result, pd.DataFrame)

    result = db.get_seqinfo("canFam6")
    assert isinstance(result, pd.DataFrame)

    result = db.get_seqinfo("GRCm38")
    assert isinstance(result, pd.DataFrame)

    with pytest.raises(ValueError):
        db.get_seqinfo("NonExistentAssembly")



from __future__ import annotations

from . import core
from .interface import AssemblyInfo

for module in [core]:
    for name, func in module.__dict__.items():
        if callable(func) and not name.startswith("_"):
            setattr(AssemblyInfo, name, func)

__all__ = ["AssemblyInfo"]


_db: AssemblyInfo | None = None

def connect() -> AssemblyInfo:
    global _db
    if _db is None:
        _db = AssemblyInfo()
    return _db




from assemblyinfo.core.assembly import Assembly
from assemblyinfo.interface import AssemblyInfo


def test_assembly_info_human():
    db = AssemblyInfo.connect()

    assembly = db.assembly_info(assembly="GRCh38")

    assert isinstance(assembly, Assembly)
    assert assembly.assembly == "GRCh38"
    assert assembly.species == "homo_sapiens"
    assert assembly.common_name == "human"

    assembly = db.assembly_info(assembly="hg38")

    assert isinstance(assembly, Assembly)
    assert assembly.assembly == "hg38"
    assert assembly.species == "homo_sapiens"
    assert assembly.common_name == "human"



import re
from typing import Any, Dict, List, Optional

import pandas as pd

__all__ = [
    "info",
    "get_db",
    "get_info",
    "get_species_info",
    "get_organism_info",
    "build_assembly_info",
    "get_version",
    "get_assembly_metadata",
    "available_assemblies",
    "available_patches",
    "available_species",
    "available_accessions",
]


def get_db(cls) -> pd.DataFrame:
    """
    Returns the AssemblyInfo database.

    This method returns the database stored in the class attribute `_data`.

    Parameters
    ----------
    cls : type
        The class object containing the AssemblyInfo data.

    Returns
    -------
    pd.DataFrame
        The database stored in the class attribute `_data`.

    Examples
    --------
    >>> AssemblyInfo.get_db()
    """

    return cls._data


def info(cls) -> str:
    """
    Display information about available entries in AssemblyInfo.

    This method prints a formatted message showing the unique values for
    species, UCSC assemblies, and NCBI assemblies available in the
    AssemblyInfo data.

    Parameters
    ----------
    cls : type
        The class object containing the AssemblyInfo data.

    Returns
    -------
    None
        This method doesn't return any value, it only prints information.

    Examples
    --------
    >>> AssemblyInfo.info()
    ```
    AssemblyInfo available entries:
        Species:
            human, mouse, rat
        Assemblies (UCSC):
            hg38, mm10, rn6
        Assemblies (NCBI):
            GRCh38, GRCm38, Rnor_6.0
    Please pick an entry and retrieve your desired data!
    ```

    Notes
    -----
    The data is accessed through the `_data` attribute of the class, which
    is expected to be a pandas DataFrame with columns 'species',
    'assembly_ucsc', and 'assembly'.
    """
    data = cls._data
    species_list = data["species"].unique().tolist()
    species_names = data["common_name"].unique().tolist()
    assemblies_ucsc = data["assembly_ucsc"].dropna().unique().tolist()
    assemblies_ncbi = data["assembly"].unique().tolist()

    msg = (
        "Genome Information:\n"
        "===================\n"
        f"Species:\n  - {', '.join(species_list)}\n\n"
        f"Common Names:\n  - {', '.join(species_names)}\n\n"
        f"Assemblies (UCSC):\n  - {', '.join(assemblies_ucsc)}\n\n"
        f"Assemblies (NCBI):\n  - {', '.join(assemblies_ncbi)}\n\n"
        "Please pick an entry and retrieve your desired data!"
    )

    return msg


def get_info(cls, key: str, value: Optional[str]) -> pd.DataFrame:
    """
    Retrieves information from the database based on the specified key and value.

    Parameters
    ----------
    key : str
        The key to filter by (e.g., 'species', 'common_name').
    value : Optional[str]
        The value to filter for the given key.

    Returns
    -------
    pd.DataFrame
        Filtered DataFrame based on the key and value.

    Raises
    ------
    ValueError
        If the value is None or the key is not found in the database.

    Examples
    --------
    >>> AssemblyInfo.get_info("species", "homo_sapiens")
    """
    if value is None:
        raise ValueError(f"ERROR! Pick a {key}: {cls._data[key].unique()}")

    local_db = cls._data.set_index(key).loc[value, :]
    return local_db


def get_version(s: List[str]) -> str:
    """
    Extracts the version number from each string in a list.

    Parameters
    ----------
    s : List[str]
        A list of strings containing version numbers.

    Returns
    -------
    Tuple[int]
        A tuple containing the version number.
    """
    match = re.search(r"(\d+)(?:\.\d+)*$", s)
    if match:
        return tuple(map(int, match.group().split(".")))
    return (0,)


def get_species_info(cls, species: Optional[str] = None) -> str:
    """
    Prints the genome information for the specified species.

    Parameters
    ----------
    species : Optional[str]
        The species name to filter by.

    Examples
    --------
    >>> AssemblyInfo.get_species_info("species", "homo_sapiens")
    ```
    Genome Information for homo_sapiens:
    Common Names:
      - human
    Assemblies (UCSC):
      - hg19, hg38, hg17, hg18, hs1
    Assemblies (NCBI):
      - GRCh37, GRCh38, NCBI35, NCBI36, T2T-CHM13
    ```
    """
    local_db = cls.get_info("species", species)
    species_names = local_db["common_name"].unique().tolist()
    assemblies_ucsc = local_db["assembly_ucsc"].dropna().unique().tolist()
    assemblies_ncbi = local_db["assembly"].unique().tolist()

    msg = (
        f"Genome Information for {species}:\n"
        "===================\n"
        f"Common Names:\n  - {', '.join(species_names)}\n\n"
        f"Assemblies (UCSC):\n  - {', '.join(assemblies_ucsc)}\n\n"
        f"Assemblies (NCBI):\n  - {', '.join(assemblies_ncbi)}\n\n"
    )

    return msg


def get_organism_info(cls, organism: Optional[str] = None) -> str:
    """
    Prints the genome information for the specified organism common name.

    Parameters
    ----------
    organism : Optional[str]
        The common name of the organism to filter by.

    Examples
    --------
    >>> AssemblyInfo.get_species_info("species", "homo_sapiens")
    ```
    Genome Information for human:
    Species:
      - homo_sapiens
    Assemblies (UCSC):
      - hg19, hg38, hg17, hg18, hs1
    Assemblies (NCBI):
      - GRCh37, GRCh38, NCBI35, NCBI36, T2T-CHM13
    ```
    """
    local_db = cls.get_info("common_name", organism)
    organism_names = local_db["species"].unique().tolist()
    assemblies_ucsc = local_db["assembly_ucsc"].dropna().unique().tolist()
    assemblies_ncbi = local_db["assembly"].unique().tolist()

    msg = (
        f"Genome Information for {organism}:\n"
        "===================\n"
        f"Species:\n  - {', '.join(organism_names)}\n\n"
        f"Assemblies (UCSC):\n  - {', '.join(assemblies_ucsc)}\n\n"
        f"Assemblies (NCBI):\n  - {', '.join(assemblies_ncbi)}\n\n"
    )

    return msg


def get_assembly_metadata(cls, assembly: Optional[str] = None) -> Dict[str, Any]:
    """
    Retrieves the detailed information for the specified assembly.

    Parameters
    ----------
    assembly : Optional[str]
        The assembly name to filter by, using either NCBI or UCSC nomenclature.

    Returns
    -------
    Dict[str, Any]
        A dictionary containing detailed information about the assembly.

    Raises
    ------
    ValueError
        If the assembly is not provided or not found in the database.

    Examples
    --------
    >>> AssemblyInfo.get_assembly_metadata("hg38")
        {
            'assembly_level': 'Chromosome',
            'assembly_method': None,
            'assembly_name': 'GRCh38',
            'assembly_type': 'haploid-with-alt-loci',
            'bioproject': 'PRJNA168',
        }
    """
    if assembly is None:
        error_msg = (
            "ERROR: You did not provide any assembly! ",
            "Pick an assembly using the NCBI nomenclature from:\n\n",
            f"{cls._data['assembly'].unique().tolist()}\n\n",
            "or the UCSC nomenclature from:\n\n",
            f"{cls._data['assembly_ucsc'].dropna().unique().tolist()}",
        )
        raise ValueError(error_msg)

    if assembly in cls._data["assembly"].tolist():
        local_db = cls._data.set_index("assembly").loc[assembly, :]
        if isinstance(local_db, pd.Series):
            local_db = local_db.to_frame().T.reset_index(names="assembly")
        else:
            local_db = local_db.reset_index(names="assembly")
        out = cls.build_assembly_info(local_db, assembly)

        return out

    elif assembly in cls._data["assembly_ucsc"].dropna().tolist():
        local_db = cls._data.set_index("assembly_ucsc").loc[assembly, :]
        if isinstance(local_db, pd.Series):
            local_db = local_db.to_frame().T.reset_index(names="assembly_ucsc")
        else:
            local_db = local_db.reset_index(names="assembly_ucsc")
        out = cls.build_assembly_info(local_db, assembly)

        return out
    else:
        error_msg = (
            f"ERROR: {assembly} not in database! ",
            "Pick an assembly using the NCBI nomenclature from:\n\n",
            f"{cls._data['assembly'].unique().tolist()}\n\n",
            "or the UCSC nomenclature from:\n\n",
            f"{cls._data['assembly_ucsc'].dropna().unique().tolist()}",
        )
        raise ValueError(error_msg)


def build_assembly_info(cls, local_db: pd.DataFrame, assembly: str) -> Dict[str, Any]:
    """
    Builds a dictionary with detailed information for the specified assembly.

    Parameters
    ----------
    local_db : pd.DataFrame
        The filtered DataFrame containing information about the assembly.
    assembly : str
        The name of the assembly.

    Returns
    -------
    Dict[str, Any]
        A dictionary containing detailed information about the assembly.

    Examples
    --------
    >>> AssemblyInfo.build_assembly_info(local_db, "hg38")
    """
    if len(local_db.patch) > 1:
        latest = sorted(local_db.patch.tolist(), key=get_version, reverse=True)[0]
        core = local_db.query(f"patch=='{latest}'").metadata.tolist()[0]
    else:
        core = local_db.query(
            f"patch=='{local_db.assembly.unique()[0]}'"
        ).metadata.tolist()[0]

    return dict(core, **{
        "species": local_db.species.unique()[0],
        "common_name": local_db.common_name.unique()[0],
        "synonyms": [local_db.assembly.unique()[0], local_db.assembly_ucsc.unique()[0]],
        "patches": local_db.patch.tolist(),
        "genbank": local_db.genbank_accession.tolist(),
        "refseq": local_db.refseq_accession.tolist(),
    })


def available_assemblies(cls, provider: Optional[str] = None) -> List[str]:
    """
    Returns the list of available assemblies.

    Parameters
    ----------
    provider : Optional[str]
        The provider to filter by ('ucsc' or 'ncbi').

    Returns
    -------
    List[str]
        A list of available assemblies.

    Raises
    ------
    ValueError
        If the provider is not 'ucsc' or 'ncbi'.

    Examples
    --------
    >>> AssemblyInfo.available_assemblies()
    ```
        ['WS144',
         'WBcel215',
         'WBcel235',
         'WS190',
         'WS195',
         ...
         ]
    ```
    """
    if not provider:
        return (
            cls._data.assembly.unique().tolist()
            + cls._data.assembly_ucsc.unique().tolist()
        )
    elif provider == "ucsc":
        return cls._data.assembly_ucsc.unique().tolist()
    elif provider == "ncbi":
        return cls._data.assembly.unique().tolist()
    else:
        raise ValueError("ERROR: provider must be either 'ucsc' or 'ncbi'!")


def available_patches(cls, assembly: Optional[str] = None) -> List[str]:
    """
    Returns the list of available patches.

    Parameters
    ----------
    assembly : Optional[str]
        The assembly name to filter by.

    Returns
    -------
    List[str]
        A list of available patches.

    Examples
    --------
    >>> AssemblyInfo.available_patches('GRCh38')
    ```
        ['GRCh38',
         'GRCh38.p1',
         'GRCh38.p2',
         'GRCh38.p3',
         'GRCh38.p4',
         'GRCh38.p5',
         'GRCh38.p6',
         'GRCh38.p7',
         'GRCh38.p8',
         'GRCh38.p9',
         'GRCh38.p10',
         'GRCh38.p11',
         'GRCh38.p12',
         'GRCh38.p13',
         'GRCh38.p14']
    ```
    """
    if not assembly:
        return cls._data.patch.unique().tolist()
    else:
        return cls._data.query(f"assembly=='{assembly}'").patch.unique().tolist()


def available_species(cls) -> List[str]:
    """
    Returns the list of available species.

    Returns
    -------
    List[str]
        A list of available species.

    Examples
    --------
    >>> AssemblyInfo.available_species()
    ```
    ['homo_sapiens', 'mus_musculus']
    ```
    """
    return cls._data.species.unique().tolist()


def available_accessions(cls, assembly: str) -> List[str]:
    """
    Returns the list of available accessions for the specified assembly.

    Parameters
    ----------
    assembly : str
        The assembly name to filter by.

    Returns
    -------
    List[str]
        A list of available accessions.

    Raises
    ------
    ValueError
        If the assembly is not provided.

    Examples
    ---------
    >>> AssemblyInfo.available_accessions('hg38')
     ```
    ['GCA_000001405.15',
     'GCA_000001405.16',
     'GCA_000001405.17',
     'GCA_000001405.18',
     'GCA_000001405.19']
    ```
    """
    if not assembly:
        raise ValueError("ERROR: you must provide an assembly!")

    if assembly in cls._data["assembly"].tolist():
        db = cls._data.query(f"assembly=='{assembly}'")
        return db["genbank_accession"].tolist() + db["refseq_accession"].tolist()
    else:
        db = cls._data.query(f"assembly_ucsc=='{assembly}'")
        return db["genbank_accession"].tolist() + db["refseq_accession"].tolist()



from typing import List

__all__ = [
    "get_genbank_accession",
    "get_refseq_accession",
    "get_patch_from_accession",
    "get_assembly_from_accession",
]


def get_genbank_accession(cls, patch: str) -> str:
    """
    Returns the GenBank accession for the specified patch.

    Parameters
    ----------
    patch : str
        The patch name to filter by.

    Returns
    -------
    str
        The GenBank accession.['GRCh38.p14']

    Raises
    ------
    ValueError
        If the patch is not provided.

    Examples
    --------
    >>> AssemblyInfo.get_genbank_accession("GRCh38.p14")
    """
    if not patch:
        raise ValueError("ERROR: you must provide a patch!")
    elif patch not in cls._data.patch.tolist():
        raise ValueError("ERROR: patch not in database!")

    return cls._data.query(f"patch=='{patch}'").genbank_accession.tolist()


def get_refseq_accession(cls, patch: str) -> str:
    """
    Returns the RefSeq accession for the specified patch.

    Parameters
    ----------
    patch : str
        The patch name to filter by.

    Returns
    -------
    str
        The RefSeq accession.

    Raises
    ------
    ValueError
        If the patch is not provided.

    Examples
    --------
    >>> AssemblyInfo.get_refseq_accession("GRCh38.p14")
    """
    if not patch:
        raise ValueError("ERROR: you must provide a patch!")
    elif patch not in cls._data.patch.tolist():
        raise ValueError("ERROR: patch not in database!")

    return cls._data.query(f"patch=='{patch}'").refseq_accession.tolist()


def get_patch_from_accession(cls, accession: str) -> List[str]:
    """
    Returns the patches for the specified accession.

    Parameters
    ----------
    accession : str
        The accession to filter by.

    Returns
    -------
    List[str]
        A list of patches.

    Raises
    ------
    ValueError
        If the accession is not provided or not found in the database.

    Examples
    --------
    >>> AssemblyInfo.get_patch_from_accession("GCA_000001405.29")
    """
    if not accession:
        raise ValueError("ERROR: you must provide an accession!")
    elif (
        accession not in cls._data.genbank_accession.dropna().tolist()
        and accession not in cls._data.refseq_accession.dropna().tolist()
    ):
        raise ValueError("ERROR: accession not in database!")

    if accession in cls._data.genbank_accession.dropna().tolist():
        return cls._data.query(f"genbank_accession=='{accession}'").patch.tolist()
    elif accession in cls._data.refseq_accession.dropna().tolist():
        return cls._data.query(f"refseq_accession=='{accession}'").patch.tolist()
    else:
        raise ValueError("ERROR: accession not in database!")


def get_assembly_from_accession(cls, accession: str) -> List[str]:
    """
    Returns the assembly names for the specified accession.

    Parameters
    ----------
    accession : str
        The accession to filter by.

    Returns
    -------
    List[str]
        A list of assembly names.

    Raises
    ------
    ValueError
        If the accession is not provided or not found in the database.

    Examples
    --------
    >>> AssemblyInfo.get_assembly_from_accession("GCA_000001405.29")
    """
    if not accession:
        raise ValueError("ERROR: you must provide an accession!")
    elif (
        accession not in cls._data.genbank_accession.dropna().tolist()
        and accession not in cls._data.refseq_accession.dropna().tolist()
    ):
        raise ValueError("ERROR: accession not in database!")

    if accession in cls._data.genbank_accession.dropna().tolist():
        return (
            cls._data.query(f"genbank_accession=='{accession}'")
            .reset_index()
            .loc[0, ["assembly", "assembly_ucsc"]]
            .tolist()
        )
    elif accession in cls._data.refseq_accession.dropna().tolist():
        return (
            cls._data.query(f"refseq_accession=='{accession}'")
            .reset_index()
            .loc[0, ["assembly", "assembly_ucsc"]]
            .tolist()
        )
    else:
        raise ValueError("ERROR: accession not in database!")



import numpy as np

NCBI = "https://ftp.ncbi.nlm.nih.gov/genomes/all"

ASSEMBLY_BLACKLIST = [
    "T2T-CHM13v0.7",
    "NCBI33",
    "NCBI34",
    "MGSCv3",
    "canfam4",
    "Release_6_plus_MT",
]
REFSEQ_BLACKLIST = [
    "GCF_000001215.2",
    "GCF_000001215.2",
    "GCF_000002285.2",
    "GCF_000002315.2",
    "GCF_000002315.6",
    "GCF_000002035.3",
]

MAP_SPECIES_NAME = {
    "homo_sapiens": "human",
    "mus_musculus": "mouse",
    "canis_lupus_familiaris": "dog",
    "caenorhabditis_elegans": "celegans",
    "drosophila_melanogaster": "fruitfly",
    "danio_rerio": "zebrafish",
    "gallus_gallus": "chicken",
    "bos_taurus": "cow",
}

HUMAN_ASSEMBLY_MAP = {
    "GRCh37": "hg19",
    "GRCh38": "hg38",
    "T2T-CHM13": "hs1",
    "NCBI34": "hg16",
    "NCBI35": "hg17",
    "NCBI36": "hg18",
}
MOUSE_ASSEMBLY_MAP = {
    "GRCm38": "mm10",
    "GRCm39": "mm39",
    "MGSCv37": "mm9",
    "MGSCv36": "mm8",
    "MGSCv35": "mm7",
    "MGSCv34": "mm6",
}
DOG_ASSEMBLY_MAP = {
    "UU_Cfam_GSD_1.0": "canFam4",
    "CanFam2.0": "canFam2",
    "CanFam3.1": "canFam3",
    "Dog10K_Boxer_Tasha": "canFam6",
    "ROS_Cfam_1.0": "ROS_Cfam_1.0",
    "UMICH_Zoey_3.1": "canFam5",
}
CELEGANS_ASSEMBLY_MAP = {
    "WS144": np.NaN,
    "WBcel215": np.NaN,
    "WBcel235": "ce11",
    "WS190": "ce6",
    "WS195": np.NaN,
    "ASM2820141v1": np.NaN,
}
DROSOPHILA_ASSEMBLY_MAP = {
    "Release_5": "dm3",
    "Release_6": "dm6",
    "Release_6_plus_ISO1_MT": "dm6",
    "Release_6_plus_MT": "dm6",
}
DANIO_ASSEMBLY_MAP = {
    "Zv8": "danRer6",
    "Zv9": "danRer7",
    "GRCz10": "danRer10",
    "GRCz11": "danRer11",
    "Zv7": "danRer5",
    "ASM3317019v1": np.NaN,
    "ASM3317019v2": np.NaN,
}
CHICKEN_ASSEMBLY_MAP = {
    "bGalGal1.mat.broiler.GRCg7b": "galGal7",
    "Gallus_gallus-2.1": "galGal3",
    "Gallus_gallus-4.0": "galGal4",
    "Gallus_gallus-5.0": "galGal5",
    "GRCg6": np.NaN,
    "GRCg6a": "galGal6",
}
COW_ASSEMBLY_MAP = {
    "ARS-UCD1.1": np.NaN,
    "ARS-UCD1.2": "bosTau9",
    "ARS-UCD1.3": np.NaN,
    "ARS-UCD2.0": np.NaN,
}
ASSEMBLY_MAP = (
    HUMAN_ASSEMBLY_MAP
    | MOUSE_ASSEMBLY_MAP
    | DOG_ASSEMBLY_MAP
    | CELEGANS_ASSEMBLY_MAP
    | DROSOPHILA_ASSEMBLY_MAP
    | DANIO_ASSEMBLY_MAP
    | CHICKEN_ASSEMBLY_MAP
    | COW_ASSEMBLY_MAP
)



import re
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
import requests
from bs4 import BeautifulSoup
from schema import (
    ASSEMBLY_BLACKLIST,
    ASSEMBLY_MAP,
    MAP_SPECIES_NAME,
    NCBI,
    REFSEQ_BLACKLIST,
)


def get_directories(url: str) -> List[str]:
    """
    Retrieves a list of directories from the given URL.

    Parameters
    ----------
    url : str
        The URL to fetch the directories from.

    Returns
    -------
    List[str]
        A list of directories found in the URL.
    """
    try:
        response = requests.get(url)
        soup = BeautifulSoup(response.text, "html.parser")
        directories = [
            a.get_text() for a in soup.find_all("a") if a.get_text().endswith("/")
        ]
    except Exception as e:
        print(f"Error fetching directories for {url}: {e!s}")

    if len(directories) == 0:
        print(f"error with {url}, 0 dirs")
    return directories


def get_formatted_paths(paths: List[str]) -> List[Tuple]:
    """
    Helper function to format the paths and extract relevant information.

    Parameters
    ----------
    paths : List[str]
        A list of paths to format.

    Returns
    -------
    List[Tuple]
        A list of tuples containing the formatted path information.
    """
    return [
        (
            "_".join(x.split("_", 2)[:2]),  # Accession
            x.split("_", 2)[-1][:-1],  # assembly complete
            x.split("_", 2)[-1][
                :-1
            ],
            f"{NCBI}/{x.split('.', 1)[0].split('_')[0]}/" +
            f"{x.split('.', 1)[0].split('_')[1][0:3]}/" +
            f"{x.split('.', 1)[0].split('_')[1][3:6]}/" +
            f"{x.split('.', 1)[0].split('_')[1][6:9]}/{x}",
        )
        for x in paths
    ]


def create_dataframe(paths: List[str], columns: List[str]) -> pd.DataFrame:
    """
    Helper function to create a DataFrame from formatted data.

    Parameters
    ----------
    paths : List[str]
        A list of paths to format.
    columns : List[str]
        A list of column names for the DataFrame.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the formatted path information.
    """
    formatted_data = get_formatted_paths(paths)
    return pd.DataFrame(formatted_data, columns=columns)


def build_db(raw_db: pd.DataFrame) -> pd.DataFrame:
    """
    Builds a DataFrame from the raw database.

    Parameters
    ----------
    raw_db : pd.DataFrame
        The raw database to build the DataFrame from.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the formatted path information.
    """
    dfs = []

    for assembly, species in zip(raw_db["assembly_name"], raw_db["species"]):
        genbank_path = raw_db.set_index("assembly_name").loc[assembly, "genbank_path"]
        refseq_path = raw_db.set_index("assembly_name").loc[assembly, "refseq_path"]

        gb = get_directories(genbank_path)
        rf = get_directories(refseq_path)

        df1_gb = create_dataframe(
            gb, ["genbank_accession", "assembly_patch", "assembly", "genbank_path"]
        )
        df1_rf = create_dataframe(
            rf, ["refseq_accession", "assembly_patch", "assembly", "refseq_path"]
        )

        df = pd.merge(df1_gb, df1_rf, on="assembly_patch", how="outer")
        df["species"] = species
        df["common_name"] = MAP_SPECIES_NAME[species]
        dfs.append(df)

    main_df = pd.concat(dfs).reset_index(drop=True)
    main_df["assembly_x"].fillna(main_df["assembly_y"].to_dict(), inplace=True)
    main_df.drop(columns=["assembly_y"], inplace=True)

    main_df = main_df[
        [
            "assembly_x",
            "assembly_patch",
            "genbank_accession",
            "refseq_accession",
            "genbank_path",
            "refseq_path",
            "species",
        ]
    ]

    main_df = main_df[~main_df["assembly_patch"].isin(ASSEMBLY_BLACKLIST)].rename(
        columns={"assembly_x": "assembly", "assembly_patch": "patch"}
    )

    main_df["assembly"] = main_df["assembly"].apply(
        lambda x: (
            x.split("v")[0]
            if x.startswith("T2T")
            else x.split(".")[0]
            if x.startswith(("GRCh38.", "GRCh37.", "GRCm38."))
            else x
        )
    )

    main_df["assembly_ucsc"] = [ASSEMBLY_MAP[a] for a in main_df["assembly"]]

    main_df = main_df[
        ~main_df["refseq_accession"].isin(REFSEQ_BLACKLIST)
    ].convert_dtypes()

    return main_df


def retrieve_file_from_url(url: str, pattern: str) -> List[str]:
    """
    Retrieves a list of files from the given URL.

    Parameters
    ----------
    url : str
        The URL to fetch the files from.

    Returns
    -------
    List[str]
        A list of files found in the URL.
    """
    try:
        response = requests.get(url)
        soup = BeautifulSoup(response.text, "html.parser")
        files = [
            a.get_text() for a in soup.find_all("a") if a.get_text().endswith(pattern)
        ]
    except Exception as e:
        print(f"Error fetching directories for {url}: {e!s}")

    if len(files) == 0:
        print(f"error with {url}, 0 dirs")
    else:
        paths = []
        for f in files:
            path = f"{url}/{f}"
            paths.append(path)
    return paths


def get_metadata_info(url: str) -> Dict[str, str]:
    """
    Reads the report file line by line until a '##' is encountered,
    then splits the read lines and builds a dictionary from them.

    Parameters
    ----------
    url : str
        The URL to fetch the metadata from.

    Returns
    -------
    Dict[str, str]
        A dictionary containing the metadata information.
    """
    file = retrieve_file_from_url(url, "report.txt")
    report_dict = {}
    try:
        response = requests.get(file[0])
        response.raise_for_status()  # Ensure the request was successful
        lines = response.text.split("\n")
        collected_lines = []

        for line in lines:
            if line.strip() == "##" or len(line[1:].strip()) == 0:
                break
            collected_lines.append(line)

        for item in collected_lines:
            item = item[1:]
            key, value = item.split(":", 1)  # Split only on the first colon
            report_dict[key.strip().lower().replace(" ", "_")] = value.strip()

    except Exception as e:
        print(f"Error reading report file from {url}: {e!s}")

    return report_dict


def get_stats_info(url: str) -> pd.DataFrame:
    """
    Reads the stats file line by line, then splits the read
    lines and builds a dictionary from them.

    Parameters
    ----------
    url : str
        The URL to fetch the stats from.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the stats information.
    """
    file = retrieve_file_from_url(url, "stats.txt")
    dfs = []
    try:
        response = requests.get(file[0])
        response.raise_for_status()  # Ensure the request was successful
        lines = response.text.split("\n")
        collected_lines = []

        for line in lines:
            if line.startswith("#"):
                continue
            collected_lines.append(line)

        for item in collected_lines:
            if item.strip():
                inner_df = pd.DataFrame(item.strip().split("\t")).T.replace(
                    "na", np.NaN
                )
                inner_df.columns = [
                    "unit-name",
                    "molecule-name",
                    "molecule-type",
                    "sequence-type",
                    "statistic",
                    "value",
                ]
                dfs.append(inner_df)

    except Exception as e:
        print(f"Error reading stats file from {url}: {e!s}")

    return pd.concat(dfs)


def get_chromosome_info(url: str) -> pd.DataFrame:
    """
    Reads the report file line by line until a '##' is encountered,
    then splits the read lines and builds a dictionary from them.

    Parameters
    ----------
    url : str
        The URL to fetch the chromosome information from.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the chromosome information.
    """
    file = retrieve_file_from_url(url, "report.txt")
    dfs = []
    try:
        response = requests.get(file[0])
        response.raise_for_status()  # Ensure the request was successful
        lines = response.text.split("\n")
        collected_lines = []

        for line in lines:
            if line.startswith("#"):
                continue
            collected_lines.append(line)

        for item in collected_lines:
            if item.strip():
                inner_df = pd.DataFrame(item.strip().split("\t")).T.replace(
                    "na", np.NaN
                )
                inner_df.columns = [
                    "ncbi",
                    "role",
                    "molecule",
                    "drop",
                    "genbank",
                    "drop1",
                    "refseq",
                    "unit",
                    "length",
                    "name",
                ]
                dfs.append(inner_df)

    except Exception as e:
        print(f"Error reading report file from {url}: {e!s}")

    return pd.concat(dfs)


def process_chromosome_info(df: pd.DataFrame) -> pd.DataFrame:
    """
    Processes the chromosome information DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame containing the chromosome information.

    Returns
    -------
    pd.DataFrame
        A processed DataFrame containing the chromosome information.
    """
    df = df.drop(columns=["drop", "drop1"]).reset_index(drop=True)
    df["role"] = [role.split("-")[0] for role in df["role"]]
    df = df[~df["role"].isin(["pseudo"])]

    df["unit"] = [
        unit.split()[0].lower().strip() if unit.startswith("P") else unit
        for unit in df["unit"]
    ]

    df["name"].fillna(
        {idx: f"chr{x}" for idx, x in enumerate(df["ncbi"])}, inplace=True
    )

    df = df.dropna(how="all", axis=1)

    alias = {
        molecule: name
        for name, molecule in list(
            zip(
                df[df["role"] == "assembled"]["name"],
                df[df["role"] == "assembled"]["molecule"],
            )
        )
    }

    df["molecule"] = [
        alias[mol] if mol is not np.NaN and mol in alias else mol
        for mol in df["molecule"]
    ]

    df = df.convert_dtypes()
    df["length"] = df["length"].astype(pd.Int64Dtype())

    return df


def insert_stat_info(df: pd.DataFrame, idx: int, path: str):
    """
    Inserts the stats information into the DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to insert the stats information into.

    idx : int
        The index of the DataFrame to insert the stats information into.

    path : str
        The path to the stats information.
    """
    stats = get_stats_info(path)
    stats = stats[stats["unit-name"].isin(["all", "Primary Assembly", "non_nuclear"])]
    stats["molecule-type"].fillna("all", inplace=True)
    stats["sequence-type"] = [s.split("-")[0] for s in stats["sequence-type"]]

    # assembly
    stats_all = stats[stats["molecule-type"] == "all"].reset_index(drop=True)

    for i in stats_all.index:
        if (
            stats_all.loc[i, "unit-name"] == "Primary Assembly"
            and stats_all.loc[i, "sequence-type"] == "all"
        ):
            stats_all.loc[i, "sequence-type"] = "primary"
            stats_all.loc[i, "statistic"] = "-".join(
                [stats_all.loc[i, "sequence-type"], stats_all.loc[i, "statistic"]]
            )
        elif stats_all.loc[i, "unit-name"] != "all":
            tmp_name = stats_all.loc[i, "unit-name"].lower().replace(" ", "_")
            stats_all.loc[i, "statistic"] = f"{tmp_name}-" + "-".join(
                [stats_all.loc[i, "sequence-type"], stats_all.loc[i, "statistic"]]
            )

    df.at[idx, "metadata"] = df.loc[idx, "metadata"] | stats_all[
        ["statistic", "value"]
    ].set_index("statistic").astype(pd.Float64Dtype()).to_dict().get("value")

    # chr
    stats_chr = stats[stats["molecule-type"] != "all"].reset_index(drop=True)
    stats_chr["statistic"] = [
        "-".join([stats_chr.loc[idx, "sequence-type"], stats_chr.loc[idx, "statistic"]])
        for idx in stats_chr.index
    ]

    for i in stats_chr.index:
        if stats_chr.loc[i, "unit-name"] != "Primary Assembly":
            name = stats_chr.loc[i, "unit-name"].lower().replace(" ", "_")
            stats_chr.loc[i, "statistic"] = f"{name}-" + stats_chr.loc[i, "statistic"]

    stats_chr = stats_chr[["molecule-name", "statistic", "value"]]
    stats_chr = stats_chr.pivot(
        index="molecule-name", columns="statistic", values="value"
    ).astype(pd.Float64Dtype())
    # drop all columns filled with None

    tmp_df = pd.DataFrame.from_records(df.loc[idx, "seqinfo"]).set_index("ncbi")
    df.at[idx, "seqinfo"] = (
        pd.merge(tmp_df, stats_chr, left_index=True, right_index=True, how="left")
        .reset_index()
        .to_dict(orient="records")
    )


def get_version(s: List[str]):
    """
    Extracts the version number from each string in a list.

    Parameters
    ----------
    s : List[str]
        A list of strings containing version numbers.

    Returns
    -------
    Tuple[int]
        A tuple containing the version number.
    """
    match = re.search(r"(\d+)(?:\.\d+)*$", s)
    if match:
        return tuple(map(int, match.group().split(".")))
    return (0,)


def builder(init_db: pd.DataFrame) -> pd.DataFrame:
    """
    Builds the database from the initial DataFrame.

    Parameters
    ----------
    init_db : pd.DataFrame
        The initial DataFrame to build the database from.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the formatted path information.
    """
    db = build_db(init_db).reset_index(drop=True)

    db["seqinfo"] = [dict] * len(db)
    db["metadata"] = [dict] * len(db)

    for idx in db.index:
        path1 = db.loc[idx, "genbank_path"]
        path2 = db.loc[idx, "refseq_path"]

        if pd.notna(path1):
            report_metadata = get_metadata_info(path1)
            chrom_df = get_chromosome_info(path1)
            chrom_df = process_chromosome_info(chrom_df)

            db.at[idx, "seqinfo"] = chrom_df.to_dict(orient="records")
            db.at[idx, "metadata"] = report_metadata
            insert_stat_info(db, idx, path1)
        else:
            report_metadata = get_metadata_info(path2)
            chrom_df = get_chromosome_info(path2)
            chrom_df = process_chromosome_info(chrom_df)

            db.at[idx, "seqinfo"] = chrom_df.to_dict(orient="records")
            db.at[idx, "metadata"] = report_metadata
            insert_stat_info(db, idx, path2)

    for p in db.groupby("assembly").patch:
        sorted_patches = sorted(p[-1].tolist(), key=get_version, reverse=True)
        if len(sorted_patches) > 1:
            patch = sorted_patches[1]
        else:
            patch = sorted_patches[0]

        db.loc[db.query(f"patch=='{patch}'").index, "version"] = "latest"

    return db
